(defproject slingshot "0.10.3"
  :description "Enhanced throw, try, leveraging Clojure's capabilities"
  :dependencies [[org.clojure/clojure "1.2.1"]]
  :java-source-path "java/src"
  :warn-on-reflection true)
